import logging

from bubuku.features.rebalance import BaseRebalanceChange
from bubuku.features.rebalance.broker import BrokerDescription
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.rebalance')


def distribute(amount: int, items: list, weight_key):
    if not items:
        return []
    items = sorted(items, key=weight_key, reverse=True)
    ceil_amount = int(amount / len(items))
    amounts = [ceil_amount for _ in range(0, len(items))]
    for idx in range(0, min(amount - sum(amounts), len(items))):
        amounts[idx] += 1
    return amounts


class DistributionMap(object):
    """
    Topic distribution map. Used to correctly balance leadership across brokers. Internal collection of candidates is a
    dict with reflection of (source_broker, target_broker) -> sorted_list_of_topics_and weights.
    """
    __slots__ = [
        '_candidates',
        '_candidates_cardinality'
    ]

    def __init__(self, brokers: iter):
        self._candidates = {}
        self._candidates_cardinality = {broker: broker.calculate_topic_cardinality() for broker in brokers}
        for source_broker in brokers:
            if not source_broker.have_extra_leaders():
                continue
            for target_broker in brokers:
                if not target_broker.have_less_leaders():
                    continue
                weight_list = []
                for topic in self._candidates_cardinality[source_broker].keys():
                    delta = self._candidates_cardinality[source_broker][topic] - \
                            self._candidates_cardinality[target_broker].get(topic, 0)
                    weight_list.append((topic, delta))
                # Now the first element is the first to rebalance.
                if weight_list:
                    self._candidates[(source_broker, target_broker)] = sorted(
                        weight_list, key=lambda x: x[1], reverse=True)

    def take_move_pair(self) -> tuple:
        """
        Selects best source_broker, target_broker, topic to move leadership. Internal data is updated as if change was
        already performed
        :return: tuple source_broker, target_broker, topic_name
        """
        top_candidate = None
        for broker_pair, weight_list in self._candidates.items():
            if not top_candidate:
                top_candidate = broker_pair
            else:
                if weight_list[0][1] > self._candidates[top_candidate][0][1]:
                    top_candidate = broker_pair
                    # Taking best topic to move
        topic, _ = self._candidates[top_candidate][0]
        self._candidates_cardinality[top_candidate[0]][topic] -= 1
        topic_exhausted = self._candidates_cardinality[top_candidate[0]][topic] == 0
        if topic_exhausted:
            for broker_pair, weight_list in self._candidates.items():
                if top_candidate == broker_pair:
                    DistributionMap._rearrange_topic_weights(weight_list, topic, None)
                elif top_candidate[0] == broker_pair[0]:
                    DistributionMap._rearrange_topic_weights(weight_list, topic, None)
                elif top_candidate[1] == broker_pair[1]:
                    DistributionMap._rearrange_topic_weights(weight_list, topic, -1)
        else:
            # Updating cardinality map
            for broker_pair, weight_list in self._candidates.items():
                if top_candidate == broker_pair:
                    DistributionMap._rearrange_topic_weights(weight_list, topic, -2)
                elif top_candidate[0] == broker_pair[0] or top_candidate[1] == broker_pair[1]:  # src or dst matches
                    DistributionMap._rearrange_topic_weights(weight_list, topic, -1)
        # Get broker objects
        return top_candidate[0], top_candidate[1], topic

    @staticmethod
    def _rearrange_topic_weights(array: list, topic: str, cardinality_change):
        old_cardinality = None
        for item in array:
            if item[0] == topic:
                old_cardinality = item[1]
                array.remove(item)
                break
        if old_cardinality is None:
            return
        if cardinality_change is None:
            return
        new_cardinality = old_cardinality + cardinality_change
        idx = 0
        for i in range(len(array) - 1, -1, -1):
            if array[i][1] >= new_cardinality:
                idx = i + 1
                break
        array.insert(idx, (topic, new_cardinality))

    def cleanup(self):
        for bp in [bp for bp, wl in self._candidates.items()
                   if not bp[0].have_extra_leaders() or not bp[1].have_less_leaders()]:
            del self._candidates[bp]


class OptimizedRebalanceChange(BaseRebalanceChange):
    _LOAD_STATE = 'load_state'
    _COMPUTE_LEADERS = 'compute_leaders'
    _COMPUTE_REPLICAS = 'compute_replicas'
    _SORT_ACTIONS = 'sort_actions'
    _BALANCE = 'balance'

    def __init__(self, zk: BukuExhibitor, broker_ids: list, empty_brokers: list, exclude_topics: list,
                 parallelism: int = 1):
        self.zk = zk
        self.all_broker_ids = sorted(int(id_) for id_ in broker_ids)
        self.broker_ids = sorted(int(id_) for id_ in broker_ids if id_ not in empty_brokers)
        self.exclude_topics = exclude_topics
        self.broker_distribution = None
        self.source_distribution = None
        self.action_queue = []
        self.state = OptimizedRebalanceChange._LOAD_STATE
        self.parallelism = parallelism

    def __str__(self):
        return 'OptimizedRebalance state={}, queue_size={}, parallelism={}'.format(
            self.state, len(self.action_queue) if self.action_queue is not None else None, self.parallelism)

    def run(self, current_actions) -> bool:
        # Stop rebalance if someone is restarting
        if self.should_be_paused(current_actions):
            _LOG.warning("Rebalance paused, because other blocking events running: {}".format(current_actions))
            return True
        if self.zk.is_rebalancing():
            return True
        new_broker_ids = sorted(int(id_) for id_ in self.zk.get_broker_ids())
        if new_broker_ids != self.all_broker_ids:
            _LOG.warning("Rebalance stopped because of broker list change from {} to {}".format(
                self.all_broker_ids, new_broker_ids))
            return False
        if self.state == OptimizedRebalanceChange._LOAD_STATE:
            self._load_data()
            self.state = OptimizedRebalanceChange._COMPUTE_LEADERS
        elif self.state == OptimizedRebalanceChange._COMPUTE_LEADERS:
            self._rebalance_leaders()
            self.state = OptimizedRebalanceChange._COMPUTE_REPLICAS
        elif self.state == OptimizedRebalanceChange._COMPUTE_REPLICAS:
            self._rebalance_replicas()
            self.state = OptimizedRebalanceChange._SORT_ACTIONS
        elif self.state == OptimizedRebalanceChange._SORT_ACTIONS:
            self.action_queue = self._sort_actions()
            self.state = OptimizedRebalanceChange._BALANCE
        elif self.state == OptimizedRebalanceChange._BALANCE:
            return not self._balance()
        return True

    def _balance(self):
        items = []
        while self.action_queue and len(items) < self.parallelism:
            items.append(self.action_queue.popitem())  # key, partition tuple
        if not items:
            return True
        data_to_rebalance = [(key[0], key[1], replicas) for key, replicas in items]
        if not self.zk.reallocate_partitions(data_to_rebalance):
            for key, replicas in items:
                self.action_queue[key] = replicas
        return False

    def _rebalance_replicas(self):
        """
        Balances replicas distribution.
        """
        # Remove duplicates
        self._remove_replica_copies()
        if not self._rebalance_replicas_template(False):
            # It may happen, that data can not be copied, because there is leader of this topic already there.
            self._rebalance_replicas_template(True)
            if not self._rebalance_replicas_template(False):
                _LOG.error('Failed to rebalance replicas. Probably because of replication factor problems. '
                           'Will just stop the process')
                raise Exception('Failed to perform replica rebalance {}, {}, {}'.format(
                    self.broker_distribution, self.broker_ids, self.action_queue))

    def _sort_actions(self):
        result = {}
        for topic_partition, source_broker_id, target_broker_id in self.action_queue:
            if topic_partition not in result:
                result[topic_partition] = list(self.source_distribution[topic_partition])
            tmp_result = result[topic_partition]
            for i in range(len(tmp_result) - 1, -1, -1):
                # Here comes small trick. Each action includes only one replacement. We are starting replacements from
                # back and perform them only once.
                if tmp_result[i] == source_broker_id:
                    tmp_result[i] = target_broker_id
                    break
        return result

    def _list_active_brokers_with_skip(self, skip_id: int) -> list:
        """
        Lists active brokers without broker with id `skip_id`
        :param skip_id: Broker id to skip
        :return:
        """
        return [b for b in self.broker_distribution.values() if
                b.broker_id != skip_id and b.broker_id in self.broker_ids]

    def _remove_replica_copies(self):
        """
        During leadership rebalance it may happen that leader for replica was placed to broker that is already have
        this partition. Before starting real rebalance it is needed to move such partitions to different brokers, even
        if it will lead to some additional steps in rebalance.
        """
        for broker in self.broker_distribution.values():
            for topic_partition in broker.list_replica_copies():
                targets = [b for b in self._list_active_brokers_with_skip(broker.broker_id)]
                moved_to = broker.move_replica(topic_partition, [t for t in targets if t.has_free_replica_slots()])
                if not moved_to:
                    moved_to = broker.move_replica(topic_partition, targets)
                if not moved_to:
                    raise Exception('Failed to move replica ' + str(topic_partition) + ', not enough replicas')
                self.action_queue.append((topic_partition, broker.broker_id, moved_to.broker_id))

    def _rebalance_replicas_template(self, force: bool):
        for broker in self.broker_distribution.values():
            to_move = broker.get_replica_overload()
            if to_move <= 0:
                continue
            targets = self._list_active_brokers_with_skip(broker.broker_id)
            if not force:
                targets = [t for t in targets if t.has_free_replica_slots()]
            for _ in range(0, to_move):
                target = False
                for topic_partition in broker.list_replicas():
                    target = broker.move_replica(topic_partition, targets)
                    if target:
                        self.action_queue.append((topic_partition, broker.broker_id, target.broker_id))
                        break
                if target is None:
                    return False
                if not force:
                    targets = [t for t in targets if t.has_free_replica_slots()]
        return True

    def _rebalance_leaders(self):
        """
        Balances leadership across active brokers.
        """
        candidates = DistributionMap(self.broker_distribution.values())
        while any([broker.have_extra_leaders() for broker in self.broker_distribution.values()]):
            # Get broker objects
            source_broker, target_broker, topic = candidates.take_move_pair()
            # Selecting best partition to move (It is better to just swap leadership, instead of copying data)
            selected_partition = None
            source_partitions = source_broker.list_partitions(topic, False)
            target_partitions = target_broker.list_partitions(topic, True)
            for partition in source_partitions:
                if partition in target_partitions:
                    selected_partition = partition
                    break
            # Failed to find optimized partition for swapping, taking any.
            selected_partition = source_partitions[0] if selected_partition is None else selected_partition
            topic_partition = (topic, selected_partition)
            # Update brokers state
            target_broker.accept_leader(source_broker, topic_partition)
            # add transfer to queue
            self.action_queue.append((topic_partition, source_broker.broker_id, target_broker.broker_id))
            # Remove empty lists
            candidates.cleanup()

    def _load_data(self):
        """
        Loads data from zk and prepares to perform rebalance. Fills in leader and replica count expectations.
        """
        self.broker_distribution = {id_: BrokerDescription(id_) for id_ in self.broker_ids}
        self.source_distribution = {(topic, partition): replicas for topic, partition, replicas in
                                    self.zk.load_partition_assignment() if topic not in self.exclude_topics}
        for topic_partition, replicas in self.source_distribution.items():
            if not replicas:
                continue
            leader = True
            for replica in replicas:
                if replica not in self.broker_distribution:
                    self.broker_distribution[replica] = BrokerDescription(replica)
                if leader:
                    self.broker_distribution[replica].add_leader(topic_partition)
                    leader = False
                else:
                    self.broker_distribution[replica].add_replica(topic_partition)
        active_brokers = [broker for id_, broker in self.broker_distribution.items() if id_ in self.broker_ids]
        total_leaders = sum(b.get_leader_count() for b in self.broker_distribution.values())

        new_leader_count = distribute(total_leaders, active_brokers, BrokerDescription.get_leader_count)
        for i in range(0, len(active_brokers)):
            active_brokers[i].set_leader_expectation(new_leader_count[i])

        # Here is one trick. Across all the code replica list does not contain leaders.
        # But in order to balance correctly (max(diff(replica+leaders) <= 1) we should add leaders and substract it
        # again
        total_replicas = sum(b.get_replica_count() for b in self.broker_distribution.values()) + sum(new_leader_count)
        new_replica_count = distribute(total_replicas, active_brokers, BrokerDescription.get_replica_count)
        for i in range(0, len(active_brokers)):
            active_brokers[i].set_replica_expectation(new_replica_count[i] - new_leader_count[i])



#!/usr/bin/env python3

import logging

import click

from instance_control.command.attach import AttachCommand
from instance_control.command.create import CreateCommand
from instance_control.command.get import GetCommand
from instance_control.command.terminate import TerminateCommand
from instance_control.command.upgrade import UpgradeCommand

_LOG = logging.getLogger('bubuku.cluster.cli')


@click.group()
def cli():
    pass


@cli.command('upgrade', help='Stop Kafka gracefully, detach EBS volume, terminate the instance and launch a new one')
@click.option('--image-version', help='Docker image version to use. By default the version from config is used. '
                                      'If provided then overrides image version from config')
@click.option('--cluster-config', default='bubuku-1.json')
@click.option('--ip', required=True)
@click.option('--user', required=True)
@click.option('--odd', required=True)
@click.option('--force', is_flag=True, default=False)
def upgrade(image_version: str, cluster_config: str, ip: str, user: str, odd: str, force: bool):
    UpgradeCommand(cluster_config, image_version, ip, user, odd, force).run()


@cli.command('attach', help='Launch instance and attach it to the existing EBS volume')
@click.option('--volume-id', required=True)
@click.option('--cluster-config', default='bubuku-1.json')
def attach(volume_id: str, cluster_config: str):
    AttachCommand(cluster_config, volume_id).run()


@cli.command('create', help='Launch instance with EBS attached')
@click.option('--cluster-config', default='bubuku-1.json')
@click.option('--availability-zone', default=None)
@click.option('--image-version', help='Docker image version to use. By default the version from config is used. '
                                      'If provided then verrides image version from config')
@click.option('--instance-count', default=1, type=int, help='Count of instances to create', show_default=True)
def create(instance_count: int, availability_zone: str, image_version: str, cluster_config: str):
    CreateCommand(cluster_config, instance_count, availability_zone, image_version).run()


@cli.command('terminate', help='Terminate instance')
@click.option('--cluster-config', default='bubuku-1.json')
@click.option('--ip', required=True)
@click.option('--user', required=True)
@click.option('--odd', required=True)
def terminate(cluster_config: str, ip: str, user: str, odd: str):
    TerminateCommand(cluster_config, ip, user, odd).run()


@cli.command('get', help='Cluster nodes overview: ip, instance id and volume id')
@click.option('--cluster-config', default='bubuku-1.json')
def get(cluster_config: str):
    GetCommand(cluster_config).run()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    cli()



import logging

_LOG = logging.getLogger('bubuku.cluster.aws.taupage')


def find_amis(ec2_resource, region: str) -> dict:
    '''
    Find latest Taupage AMI for the region
    '''
    _LOG.info('Finding latest Taupage AMI in %s..', region)

    filters = [{'Name': 'name', 'Values': ['*Taupage-AMI-*']},
               {'Name': 'is-public', 'Values': ['false']},
               {'Name': 'state', 'Values': ['available']},
               {'Name': 'root-device-type', 'Values': ['ebs']}]
    images = list(ec2_resource.images.filter(Filters=filters))
    if not images:
        raise Exception('No Taupage AMI found')
    most_recent_image = sorted(images, key=lambda i: i.name)[-1]

    _LOG.info('The latest AMI is %s', most_recent_image)

    return most_recent_image


def generate_user_data(cluster_config: dict) -> str:
    '''
    Generate Taupage user data to start a Bubuku node
    http://docs.stups.io/en/latest/components/taupage.html
    '''
    _LOG.info('Generating Taupage user data')

    environment = cluster_config['environment']
    data = {'runtime': 'Docker',
            'source': cluster_config['docker_image'],
            'application_id': cluster_config['application_id'],
            'application_version': cluster_config['image_version'],
            'networking': 'host',
            'ports': {'9092': '9092',
                      '8004': '8004',
                      '8080': '8080',
                      '8778': '8778',
                      environment['health_port']: environment['health_port']},
            'enhanced_cloudwatch_metrics': True,
            'application_logrotate_size': '100M',
            'application_logrotate_rotate': 4,
            'environment': {
                'CLUSTER_NAME': cluster_config['cluster_name'],
                'ZOOKEEPER_STACK_NAME': environment['zookeeper_stack_name'],
                'ZOOKEEPER_PREFIX': environment['zookeeper_prefix'],
                'BUBUKU_MODE': environment['bubuku_mode'],
                'BUKU_FEATURES': environment['buku_features'],
                'JMX_PORT': environment['jmx_port'],
                'KAFKA_HEAP_OPTS': environment['kafka_heap_opts'],
                'HEALTH_PORT': environment['health_port'],
                'STARTUP_TIMEOUT_TYPE': environment['startup_timeout_type'],
                'STARTUP_TIMEOUT_INITIAL': environment.get('startup_timeout_initial', ),
                'STARTUP_TIMEOUT_STEP': environment.get('startup_timeout_step')
            },
            'volumes': {
                'ebs': {
                    '/dev/xvdk': None
                }
            },
            'mounts': {
                '/data': {
                    'partition': '/dev/xvdk',
                    'filesystem': 'ext4',
                    'options': 'noatime,nodiratime,nobarrier'
                }
            },
            'scalyr_account_key': cluster_config['scalyr_key']
            }

    _LOG.info('Generated data:')
    for k, v in data.items():
        if k == 'environment':
            _LOG.info('--- environment variables')
            for ek, ev in environment.items():
                _LOG.info('%s=%s', ek, ev)
            _LOG.info('--- environment variables')
        else:
            _LOG.info('%s=%s', k, v)

    return data



class _TopicPartitions(object):
    __slots__ = (
        '_items',
        '_expectation',
        '_has_free_slots',
        '_topic_partitions'
    )

    def __init__(self):
        self._items = set()
        self._expectation = 0
        self._has_free_slots = None
        self._topic_partitions = {}

    def __str__(self):
        return 'TP, expectation: {}, items: {}'.format(self._expectation, self._items)

    def get_item_count(self):
        return len(self._items)

    def contains(self, item):
        return item in self._items

    def add(self, item):
        self._items.add(item)
        topic, partition = item
        if topic not in self._topic_partitions:
            self._topic_partitions[topic] = []
        self._topic_partitions[topic].append(partition)
        self._has_free_slots = None

    def remove(self, item):
        self._items.remove(item)
        topic, partition = item
        self._topic_partitions[topic].remove(partition)
        self._has_free_slots = None

    def iterate_items(self):
        return self._items.__iter__()

    def get_topic_partitions(self, topic):
        return self._topic_partitions.get(topic, [])

    def get_expectation(self):
        return self._expectation

    def set_expectation(self, expectation):
        self._has_free_slots = None
        self._expectation = expectation

    def calculate_cardinality(self):
        return {k: len(v) for k, v in self._topic_partitions.items()}

    def has_free_slots(self):
        if self._has_free_slots is None:
            self._has_free_slots = self.get_item_count() < self.get_expectation()
        return self._has_free_slots


class BrokerDescription(object):
    __slots__ = (
        '_broker_id',
        '_leaders',
        '_replicas',
    )

    def __init__(self, broker_id: int):
        self._broker_id = broker_id
        self._leaders = _TopicPartitions()
        self._replicas = _TopicPartitions()

    @property
    def broker_id(self):
        return self._broker_id

    def __str__(self):
        return 'BrokerDescription(id={}, leaders={}, replicas={})'.format(
            self._broker_id, self._leaders, self._replicas)

    def set_leader_expectation(self, leader_count: int):
        self._leaders.set_expectation(leader_count)

    def set_replica_expectation(self, replica_count: int):
        self._replicas.set_expectation(replica_count)

    def add_leader(self, topic_partition: tuple):
        self._leaders.add(topic_partition)

    def add_replica(self, topic_partition: tuple):
        self._replicas.add(topic_partition)

    def get_leader_count(self):
        return self._leaders.get_item_count()

    def get_replica_count(self):
        return self._replicas.get_item_count()

    def get_replica_overload(self):
        return self._replicas.get_item_count() - self._replicas.get_expectation()

    def has_free_replica_slots(self):
        return self._replicas.has_free_slots()

    def have_extra_leaders(self):
        return self._leaders.get_expectation() < self._leaders.get_item_count()

    def have_less_leaders(self):
        return self._leaders.get_expectation() > self._leaders.get_item_count()

    def get_expected_leaders(self):
        return self._leaders.get_expectation()

    def accept_leader(self, source_broker, topic_partition: tuple):
        """
        Moves topic_partition from source_broker to self broker.
        :param source_broker: Broker to take topic_partition from.
        :param topic_partition: topic and partition tuple to take.
        """
        self._leaders.add(topic_partition)
        source_broker._leaders.remove(topic_partition)

    def _accept_replica(self, source_broker, topic_partition: tuple):
        # Already a leader for this partition
        if self._leaders.contains(topic_partition):
            return False
        # Already a replica for this partition
        if self._replicas.contains(topic_partition):
            return False
        self._replicas.add(topic_partition)
        source_broker._replicas.remove(topic_partition)
        return True

    def move_replica(self, topic_partition: tuple, broker_list: list):
        """
        Moves replica topic_partition to some broker from broker_list.
        :param topic_partition: Topic and partition to move
        :param broker_list: List of brokers (BrokerDescription) to move to
        :return: Broker, to which partition was moved
        """
        for target in broker_list:
            if target._accept_replica(self, topic_partition):
                return target
        return None

    def list_replica_copies(self):
        return list([tp for tp in self._replicas.iterate_items() if self._leaders.contains(tp)])

    def list_partitions(self, topic: str, replica: bool):
        return (self._replicas if replica else self._leaders).get_topic_partitions(topic)

    def list_replicas(self):
        return self._replicas.iterate_items()

    def calculate_topic_cardinality(self):
        """
        Calculates 'topic to leader count' dictionary on this broker.
        For example, topic t0 have partitions 0, 1, 2, 3. If leaders for partitions 0, 3 are located on this broker
         than return value will contain mapping t0->2 (there are 2 leaders for topic t0 on this broker)
        :return: Dictionary with leaders count per topic for this broker.
        """
        return self._leaders.calculate_cardinality()



import base64
import logging

from instance_control import config
from instance_control import node
from instance_control import piu
from instance_control import volume
from instance_control.aws import AWSResources
from instance_control.aws.ec2_node import EC2
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.upgrade')


class UpgradeCommand(Command):
    def __init__(self, cluster_config_path: str, image_version: str, ip: str, user: str, odd: str, force: bool):
        super().__init__(cluster_config_path)
        self.image_version = image_version
        self.ip = ip
        self.user = user
        self.odd = odd
        self.force = force

    def alter_config(self):
        if self.image_version:
            self.cluster_config['image_version'] = self.image_version

    def execute(self):
        aws_ = AWSResources(region=self.cluster_config['region'])

        instance = node.get_instance_by_ip(aws_.ec2_resource, self.cluster_config['cluster_name'], self.ip)
        if not self.force and self.cluster_config['image_version'] == get_image_version(instance):
            raise Exception("Current running Bubuku version is the same as provided ({}), stopping upgrade".format(
                self.cluster_config['image_version']))

        _LOG.info('Searching for instance %s volumes', instance.instance_id)
        volumes = aws_.ec2_client.describe_instance_attribute(InstanceId=instance.instance_id,
                                                              Attribute='blockDeviceMapping')
        data_volume = next(v for v in volumes['BlockDeviceMappings'] if v['DeviceName'] == '/dev/xvdk')
        data_volume_id = data_volume['Ebs']['VolumeId']

        piu.stop_taupage(self.ip, self.user, self.odd)

        _LOG.info('Creating tag:Name=%s for %s', config.KAFKA_LOGS_EBS, data_volume_id)
        vol = aws_.ec2_resource.Volume(data_volume_id)
        vol.create_tags(Tags=[{'Key': 'Name', 'Value': config.KAFKA_LOGS_EBS}])
        _LOG.info('Detaching %s from %s', data_volume_id, instance.instance_id)

        aws_.ec2_client.detach_volume(VolumeId=data_volume_id, Force=False)

        node.terminate(aws_, self.cluster_config['cluster_name'], instance)
        self.cluster_config['availability_zone'] = vol.availability_zone
        self.cluster_config['create_ebs'] = False

        ec2 = EC2(aws_)

        ec2.create(self.cluster_config, 1)
        # volumes are going to be attached by taupage
        volume.wait_volumes_attached(aws_)


def get_image_version(instance):
    response = instance.describe_attribute(Attribute='userData')
    env = base64.b64decode(response['UserData']['Value']).decode('UTF-8')
    variables = env.split("\n")
    for v in variables:
        if v.startswith("application_version"):
            return v.split(": ", 1)[1]


def check_current_image_version(instance, provided_image_version) -> bool:
    _LOG.info('Checking running Bubuku version on %s', instance.instance_id)
    running_version = get_image_version(instance)
    return running_version == provided_image_version



import logging

import boto3
from botocore.config import Config


class AWSResources(object):
    def __init__(self, region, retries=100):
        boto3.set_stream_logger('boto3', logging.INFO)
        self.session = boto3.Session()
        self.region = region
        self.retries = retries
        self._ec2_client = None
        self._ec2_resource = None
        self._cloudwatch_client = None
        self._iam_client = None

    @property
    def ec2_client(self):
        if not self._ec2_client:
            self._ec2_client = self.session.client(
                'ec2',
                region_name=self.region,
                config=Config(retries={'max_attempts': self.retries}))
        return self._ec2_client

    @property
    def ec2_resource(self):
        if not self._ec2_resource:
            self._ec2_resource = self.session.resource(
                'ec2',
                region_name=self.region,
                config=Config(retries={'max_attempts': self.retries}))
        return self._ec2_resource

    @property
    def cloudwatch_client(self):
        if not self._cloudwatch_client:
            self._cloudwatch_client = self.session.client(
                'cloudwatch',
                region_name=self.region,
                config=Config(retries={'max_attempts': self.retries}))
        return self._cloudwatch_client

    @property
    def iam_client(self):
        if not self._iam_client:
            self._iam_client = self.session.client('iam', config=Config(retries={'max_attempts': self.retries}))
        return self._iam_client



import json
import logging
from time import time, sleep

from bubuku.config import KafkaProperties
from bubuku.id_generator import BrokerIdGenerator
from bubuku.process import KafkaProcess
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.broker')


class LeaderElectionInProgress(Exception):
    pass


class StartupTimeout(object):
    def is_timed_out(self, seconds: float) -> bool:
        raise Exception('Not supported')

    def on_timeout_fail(self):
        raise Exception('Not supported')

    @staticmethod
    def build(props: dict):
        type_ = props.get('type', 'linear')
        if type_ == 'linear':
            return LinearTimeout(float(props.get('initial', 300)), float(props.get('step', '60')))
        elif type_ == 'progressive':
            return ProgressiveTimeout(float(props.get('initial', 300)), float(props.get('step', '0.5')))
        elif type_ == 'none':
            return NoTimeout()
        else:
            raise NotImplementedError('Startup timeout type {} is not valid'.format(type_))


class ProgressiveTimeout(StartupTimeout):
    def __init__(self, initial: float, scale: float):
        self.initial = initial
        self.timeout = initial
        self.scale = scale

    def is_timed_out(self, seconds: float) -> bool:
        return seconds > self.timeout

    def on_timeout_fail(self):
        self.timeout = self.timeout * (1 + self.scale)

    def __str__(self):
        return 'Progressive, initial={}, scale={}, current={}'.format(self.initial, self.scale, self.timeout)


class LinearTimeout(StartupTimeout):
    def __init__(self, initial: float, step: float):
        self.initial = initial
        self.timeout = initial
        self.step = step

    def is_timed_out(self, seconds: float) -> bool:
        return seconds > self.timeout

    def on_timeout_fail(self):
        self.timeout += self.step

    def __str__(self):
        return 'Linear, initial={}, step={}, current={}'.format(self.initial, self.step, self.timeout)


class NoTimeout(StartupTimeout):
    def is_timed_out(self, seconds: float) -> bool:
        return False

    def on_timeout_fail(self):
        pass


class BrokerManager(object):
    def __init__(self, process: KafkaProcess, exhibitor: BukuExhibitor,
                 id_manager: BrokerIdGenerator, kafka_properties: KafkaProperties, timeout: StartupTimeout):
        self.id_manager = id_manager
        self.exhibitor = exhibitor
        self.kafka_properties = kafka_properties
        self.process = process
        self.timeout = timeout

    def is_running_and_registered(self):
        if not self.process.is_running():
            return False
        return self.id_manager.is_registered()

    def stop_kafka_process(self):
        if self.process.is_running():
            self.process.stop_and_wait()
        self._wait_for_zk_absence()

    def _is_clean_election(self):
        value = self.kafka_properties.get_property('unclean.leader.election.enable')
        return value == 'false'

    def has_leadership(self):
        """
        Says if this broker is still a leader for partitions
        :return: True, if broker is a leader for some partitions.
        """
        broker_id = self.id_manager.get_broker_id()
        if not broker_id:
            return False
        return not self._is_leadership_transferred(dead_broker_ids=[broker_id])

    def _wait_for_zk_absence(self):
        try:
            while self.id_manager.is_registered():
                sleep(1)
        except Exception as e:
            _LOG.error('Failed to wait until broker id absence in zk', exc_info=e)

    def get_zk_connect_string(self):
        return self.kafka_properties.get_property('zookeeper.connect')

    def start_kafka_process(self, zookeeper_address):
        """
        Starts kafka using zookeeper address provided.
        :param zookeeper_address: Address to use for kafka
        :raise LeaderElectionInProgress: raised when broker can not be started because leader election is in progress
        """
        if not self.process.is_running():
            if not self._is_leadership_transferred(active_broker_ids=self.exhibitor.get_broker_ids()):
                raise LeaderElectionInProgress()

            _LOG.info('Using ZK address: {}'.format(zookeeper_address))
            self.kafka_properties.set_property('zookeeper.connect', zookeeper_address)

            self.kafka_properties.dump()

            _LOG.info('Staring kafka process')
            self.process.start(self.kafka_properties.settings_file)

            _LOG.info('Waiting for kafka to start with timeout {}'.format(self.timeout))
            start = time()
            while self.process.is_running():
                if self.id_manager.is_registered():
                    break
                if self.timeout.is_timed_out(time() - start):
                    self.timeout.on_timeout_fail()
                    break
                sleep(1)
            if not self.process.is_running() or not self.id_manager.is_registered():
                _LOG.error(
                    'Failed to wait for broker to start up, probably will kill, next timeout is'.format(self.timeout))

    def _is_leadership_transferred(self, active_broker_ids=None, dead_broker_ids=None):
        _LOG.info('Checking if leadership is transferred: active_broker_ids={}, dead_broker_ids={}'.format(
            active_broker_ids, dead_broker_ids))
        if self._is_clean_election():
            for topic, partition, state in self.exhibitor.load_partition_states():
                leader = str(state['leader'])
                if active_broker_ids and leader not in active_broker_ids:
                    if any(str(x) in active_broker_ids for x in state.get('isr', [])):
                        _LOG.warning(
                            'Leadership is not transferred for {} {} ({}, brokers: {})'.format(
                                topic, partition, json.dumps(state), active_broker_ids))
                        return False
                    else:
                        _LOG.warning('No single isr available for {}, {}, state: {}, skipping check for that'.format(
                            topic, partition, json.dumps(state)))
                if dead_broker_ids and leader in dead_broker_ids:
                    _LOG.warning('Leadership is not transferred for {} {}, {} (dead list: {})'.format(
                        topic, partition, json.dumps(state), dead_broker_ids))
                    return False

        return True



import json
import logging
import uuid
from functools import partial

import boto3
import requests

from bubuku.config import Config, KafkaProperties
from bubuku.id_generator import BrokerIdGenerator
from bubuku.zookeeper import BukuExhibitor, AddressListProvider
from bubuku.zookeeper.exhibitor import ExhibitorAddressProvider

_LOG = logging.getLogger('bubuku.amazon')


class EnvProvider(object):
    def get_id(self) -> str:
        raise NotImplementedError('Not implemented')

    def get_address_provider(self):
        raise NotImplementedError('Not implemented')

    def create_broker_id_manager(self, zk: BukuExhibitor, kafka_props: KafkaProperties):
        raise NotImplementedError('Not implemented')

    def get_rack(self):
        raise NotImplementedError('Not implemented')

    @staticmethod
    def create_env_provider(config: Config):
        if config.mode == 'amazon':
            return AmazonEnvProvider(config)
        elif config.mode == 'local':
            return LocalEnvProvider()
        else:
            raise NotImplementedError('Configuration mode "{}" is not supported'.format(config.mode))


class AmazonEnvProvider(EnvProvider):
    def __init__(self, config: Config):
        self.aws_addr = '169.254.169.254'
        self.config = config
        self.ip_address = None
        self._document = None

    def _get_document(self) -> dict:
        if not self._document:
            self._document = requests.get(
                'http://{}/latest/dynamic/instance-identity/document'.format(self.aws_addr),
                timeout=5).json()
            _LOG.info("Amazon specific information loaded from AWS: {}".format(json.dumps(self._document, indent=2)))
        return self._document

    def get_id(self) -> str:
        if not self.ip_address:
            self.ip_address = self._get_document()['privateIp']
        return self.ip_address

    def get_rack(self):
        return self._get_document()['availabilityZone']

    def _load_instance_ips(self, lb_name: str):
        region = self._get_document()['region']

        private_ips = []

        elb = boto3.client('elb', region_name=region)
        ec2 = boto3.client('ec2', region_name=region)

        response = elb.describe_instance_health(LoadBalancerName=lb_name)

        for instance in response['InstanceStates']:
            if instance['State'] == 'InService':
                private_ips.append(ec2.describe_instances(
                    InstanceIds=[instance['InstanceId']])['Reservations'][0]['Instances'][0]['PrivateIpAddress'])

        _LOG.info("Ip addresses for {} are: {}".format(lb_name, private_ips))
        return private_ips

    def get_address_provider(self):
        return ExhibitorAddressProvider(partial(self._load_instance_ips, self.config.zk_stack_name))

    def create_broker_id_manager(self, zk: BukuExhibitor, kafka_props: KafkaProperties):
        return BrokerIdGenerator(zk, kafka_props)


class _LocalAddressProvider(AddressListProvider):
    def get_latest_address(self) -> (list, int):
        return ('zookeeper',), 2181


class LocalEnvProvider(EnvProvider):
    unique_id = str(uuid.uuid4())

    def get_id(self) -> str:
        return self.unique_id

    def get_address_provider(self):
        return _LocalAddressProvider()

    def get_rack(self):
        return None

    def create_broker_id_manager(self, zk: BukuExhibitor, kafka_props: KafkaProperties):
        return BrokerIdGenerator(zk, kafka_props)



import subprocess


class KafkaProcess(object):
    def __init__(self, kafka_dir: str):
        self.process = None
        self.kafka_dir = kafka_dir

    def start(self, settings_file):
        if self.is_running():
            raise Exception('Kafka process already started')
        self.process = subprocess.Popen([self.kafka_dir + "/bin/kafka-server-start.sh", settings_file])

    def is_running(self) -> bool:
        if self.process:
            self.process.poll()
            return self.process.returncode is None
        return False

    def stop_and_wait(self):
        if self.process is None:
            raise Exception('Process was not started')
        self.process.terminate()
        self.process.wait()
        self.process = None



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Change, Check
from bubuku.features.restart_on_zk_change import RestartBrokerChange
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.restart_if_dead')


class CheckBrokerStopped(Check):
    def __init__(self, broker: BrokerManager, zk: BukuExhibitor):
        super().__init__()
        self.broker = broker
        self.zk = zk
        self.need_check = True

    def check(self) -> Change:
        if not self.need_check:
            return None
        if self.broker.is_running_and_registered():
            return None
        _LOG.info('Oops! Broker is dead, triggering restart')
        self.need_check = False

        # Do not start if broker is running and registered
        def _cancel_if():
            return self.broker.is_running_and_registered()

        return RestartBrokerChange(self.zk, self.broker, _cancel_if, self.on_check_removed)

    def on_check_removed(self):
        self.need_check = True

    def __str__(self):
        return 'CheckBrokerStopped'



import logging
from typing import List

from instance_control.aws import AWSResources
from instance_control.command import Command
from instance_control.command.upgrade import get_image_version

_LOG = logging.getLogger('bubuku.cluster.command.get')


def _print_table(items: List[dict]):
    headers = []
    for i in items:
        headers.extend(i.keys())
    headers = sorted(set(headers))
    lengths = {h: len(h) for h in headers}
    for d in items:
        for k, v in d.items():
            v = str(v)
            if lengths[k] < len(v):
                lengths[k] = len(v)
    format_ = ''
    for idx in range(0, len(headers)):
        format_ += '{' + str(idx) + ':' + str(lengths[headers[idx]]) + '} '
    print(format_.format(*headers))
    for item in items:
        print(format_.format(*[item.get(h) for h in headers]))


class GetCommand(Command):
    def __init__(self, cluster_config_path: str):
        super().__init__(cluster_config_path)

    def execute(self):
        aws_ = AWSResources(region=self.cluster_config['region'])
        instances = list(aws_.ec2_resource.instances.filter(Filters=[
            {'Name': 'instance-state-name', 'Values': ['running', 'pending']},
            {'Name': 'tag:Name', 'Values': [self.cluster_config['cluster_name']]}]))

        if len(instances) == 0:
            _LOG.info('Cluster has no running instances')
            return

        data = []
        for instance in instances:
            data.append({
                '1 Ip': instance.private_ip_address,
                '2 Id': instance.instance_id,
                '3 Volume': [x['Ebs']['VolumeId'] for x in instance.block_device_mappings if
                             x['DeviceName'] == '/dev/xvdk'][0],
                '4 Availability zone': instance.placement['AvailabilityZone'],
                '5 Image Version': get_image_version(instance)
            })

        _print_table(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import inspect
import os
import sys

import setuptools
from setuptools import setup
from setuptools.command.test import test
from distutils.core import Command

if sys.version_info < (3, 4, 0):
    sys.stderr.write('FATAL: Bubuku needs to be run with Python 3.4+\n')
    sys.exit(1)

__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))


def read_version(package):
    with open(os.path.join(package, '__init__.py'), 'r') as fd:
        for line in fd:
            if line.startswith('__version__ = '):
                return line.split()[-1].strip().strip("'")


NAME = 'bubuku'
MAIN_PACKAGE = 'bubuku'
VERSION = read_version(MAIN_PACKAGE)
DESCRIPTION = 'AWS support for kafka broker'
LICENSE = 'Apache License 2.0'
URL = 'https://github.com/zalando-incubator/bubuku'
AUTHOR = 'Dmitry Sorokin'
EMAIL = 'dmitriy.sorokin@zalando.de'
KEYWORDS = 'aws kafka supervisor'

# Add here all kinds of additional classifiers as defined under
# https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: Implementation :: CPython',
]

CONSOLE_SCRIPTS = [
    'bubuku-daemon = bubuku.daemon:main',
    'bubuku-cli = bubuku.cli:cli'
]


class DockerUpCommand(Command):
    description = "Start up docker compose with 3 bubuku and 1 zookeeper instances"
    user_options = [
        ('bubuku-scale=', None, 'Specify number of bubuku instances')
    ]

    def initialize_options(self):
        self.bubuku_scale = 3

    def finalize_options(self):
        pass

    def run(self):
        os.system('docker-compose up -d --build && docker-compose scale bubuku=' + str(self.bubuku_scale))


class DockerDownCommand(Command):
    description = "Stop docker compose"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('docker-compose down')


class PyTest(test):
    def run_tests(self):
        try:
            import pytest
        except:
            raise RuntimeError('py.test is not installed, run: pip install pytest')
        params = {'args': self.test_args}
        errno = pytest.main(**params)
        sys.exit(errno)


def read(fname):
    with open(os.path.join(__location__, fname)) as f:
        return f.read()


def setup_package():
    command_options = {'test': {'test_suite': ('setup.py', 'tests')}}

    setup(
        name=NAME,
        version=VERSION,
        url=URL,
        description=DESCRIPTION,
        author=AUTHOR,
        author_email=EMAIL,
        license=LICENSE,
        keywords=KEYWORDS,
        classifiers=CLASSIFIERS,
        test_suite='tests',
        packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
        install_requires=[req for req in read('requirements.txt').split('\\n') if req != ''],
        cmdclass={'test': PyTest, 'docker_up': DockerUpCommand, 'docker_down': DockerDownCommand},
        tests_require=['pytest-cov', 'pytest'],
        command_options=command_options,
        entry_points={
            'console_scripts': CONSOLE_SCRIPTS,
        },
    )

if __name__ == '__main__':
    setup_package()
